"""
Callbacks for model building process in SmolModels.

This module defines callback interfaces that let users hook into various stages
of the model building process, allowing for custom logging, tracking, visualization,
or other operations to be performed at key points.
"""

import logging
from abc import ABC
from dataclasses import dataclass
from typing import Optional, Type, Dict

from pydantic import BaseModel

from smolmodels.internal.common.datasets.interface import TabularConvertible
from smolmodels.internal.models.entities.node import Node

logger = logging.getLogger(__name__)


@dataclass
class BuildStateInfo:
    """
    Consolidated information about model build state at any point in the process.

    This class combines all information available during different stages of the model building
    process (start, end, iteration start, iteration end) into a single structure.
    """

    # Common identification fields
    intent: str
    """The natural language description of the model's intent."""

    provider: str
    """The provider (LLM) used for generating the model."""

    # Schema fields
    input_schema: Optional[Type[BaseModel]] = None
    """The input schema for the model."""

    output_schema: Optional[Type[BaseModel]] = None
    """The output schema for the model."""

    run_timeout: Optional[int] = None
    """Maximum time in seconds for each individual training run."""

    max_iterations: Optional[int] = None
    """Maximum number of iterations for the model building process."""

    timeout: Optional[int] = None
    """Maximum total time in seconds for the entire model building process."""

    # Iteration fields
    iteration: int = 0
    """Current iteration number (0-indexed)."""

    # Dataset fields
    datasets: Optional[Dict[str, TabularConvertible]] = None

    # Current node being evaluated
    node: Optional[Node] = None
    """The solution node being evaluated in the current iteration."""


class Callback(ABC):
    """
    Abstract base class for callbacks during model building.

    Callbacks allow running custom code at various stages of the model building process.
    Subclass this and implement the methods you need for your specific use case.
    """

    def on_build_start(self, info: BuildStateInfo) -> None:
        """
        Called when the model building process starts.
        """
        pass

    def on_build_end(self, info: BuildStateInfo) -> None:
        """
        Called when the model building process ends.
        """
        pass

    def on_iteration_start(self, info: BuildStateInfo) -> None:
        """
        Called at the start of each model building iteration.
        """
        pass

    def on_iteration_end(self, info: BuildStateInfo) -> None:
        """
        Called at the end of each model building iteration.
        """
        pass


# At the end of callbacks.py
from smolmodels.internal.models.callbacks.mlflow import MLFlowCallback

__all__ = [
    "Callback",
    "BuildStateInfo",
    "MLFlowCallback",
]
